## コードカバレッジ

あるタイミングでアプリケーションのどの箇所がテストされていて、どの箇所がテストされていないのか確認したくなると思います。
そのような状況に[コードカバレッジ](http://ja.wikipedia.org/w/index.php?title=%E3%82%B3%E3%83%BC%E3%83%89%E3%82%AB%E3%83%90%E3%83%AC%E3%83%83%E3%82%B8)が使われます。
カバレッジレポートを収集するためにテストを実行すると、そのテストによって実行されたすべてのクラス、メソッド、そして行の統計情報を取得します。
スクリプトの全行数と通過した行数との比率が、主要なカバレッジ基準となります。理想論では100%のカバレッジを取得する必要がありますが、実際には80%で十分でしょう。なぜなら、100%のカバレッジであったとしても致命的なエラーやクラッシュを防ぐことができないからです。

**Codeceptionはバージョン1.5からコードカバレッジ用のツールを備えています。カバレッジを収集するためには `xdebug` が必要となります。**.

![コードカバレッジの例](http://codeception.com/images/coverage.png)

カバレッジデータはローカルとリモートの両方のテストのために手動で収集することができます。
リモートのテストは、異なるノード上またはローカルのWebサーバーを利用して実行可能です。
SeleniumやPhpBrowserによるテストはコードカバレッジを収集するのは難しいように見えるかもしれませんが、Codeceptionはローカルと同じようにリモートのコードカバレッジをサポートしています。

### 設定

コードカバレッジを有効にするためには次の内容をグローバル設定ファイル `codeception.yml` に設定します。

``` yaml
coverage:
    enabled: true
```

いまはこれで大丈夫ですが、最終的なカバレッジレポートにはどのファイルが表示されるべきなのでしょうか？次のように、ブラックリストとホワイトリストのフィルタを提供することで、ファイルをフィルタリングすることができます。

``` yaml
coverage:
    whitelist:
        include:
            - app/*
        exclude:
            - app/cache/*
    blacklist:
        include:
            - app/controllers/*
        exclude:
            - app/cache/CacheProvider.php

```
ホワイトリストとブラックリストはなんでしょうか？

* **whitelist** 通過しなかったとしてもレポートに含まれるべきファイルの一覧です
* **blacklist** 通過したとしてもレポートから除外すべきファイルの一覧です

include/excludeセクションにファイルまたはディレクトリの配列を渡します。'\*' で終わるパスはディレクトリにマッチします。
また、'\*' マスクをファイル名に使用することができます。たとえば、`app/models/*Model.php` はすべてのモデルにマッチします。

There is a shortcut if you don't need that complex filters:
複雑なフィルタが必要ない場合のショートカットがあります。

``` yaml
coverage:
    enabled: true
    include:
        - app/*
    exclude:
        - app/cache/*
```

この `include` と `exclude` オプションはホワイトリストにファイルを追加または削除するためのものです。

以上のすべての設定は各テストスイートの設定ファイルにて再定義することができます。

## ローカルコードカバレッジ

基本的なコードカバレッジは機能テストと単体テストで収集することができます。
もし上に書いた設定を行っているのであれば準備完了です。
必要なことはCodeceptionを `--coverage` オプションとともに実行することです。

CloverのXMLレポートや良い感じのHTMLレポートを生成するためには、`--coverage-xml` と `--coverage-html` のオプションについても追加します。

``` yaml
codecept run --coverage --coverage-xml --coverage-html
```

XMLとHTMLレポートは `_output` ディレクトリに格納されます。レポートを確認するための一番の方法は、`tests/_output/coverage` ディレクトリから `index.html` をブラウザで開くことです。
CloverのXMLレポートはPHPStormのようなIDEやJenkinsのような継続的インテグレーションサーバーで利用されます。

## リモートコードカバレッジ

### ローカルサーバー

もしアプリケーションをウェブサーバー（Apache, Nginx, PHP WebServer）を介して実行している場合、テストされたコードに直接アクセスすることができないため、
カバレッジを収集することは簡単なことではありません。同様のことが、異なるノード上でテストされたスクリプトに対しても言えます。
これらのコードにアクセスするためには `xdebug` を `remote_enable` オプションを有効にしてインストールする必要があります。
Codeceptionもまた、アプリケーションとやり取りするためのちょっとしたスパイを必要とします。
スタンドアローンでアプリケーションを実行する際に、テストされているかどうかに関わらず、コードカバレッジ情報を収集するために小さなファイルをインクルードしてください。

このファイルは `c3.php` と呼ばれ、[GitHubから利用可能](https://github.com/Codeception/c3) です。
`c3.php` をダウンロードし、アプリケーションのコントローラの最初でインクルードしてください。
特別なヘッダー情報を送信することにより、Codeceptionはアプリケーションに対して、いつコードカバレッジの収集を開始し、いつ終了するか命令します。
テストスイートが終了した後、レポートが出力され、Codeceptionはアプリケーションからそれを取得します。

インストール方法は [readme file](https://github.com/Codeception/c3) の説明に従ってください。

`c3` への接続のために、CodeceptionはPhpBrowserやWebDriverモジュールのURL設定を利用しますが、
`c3.php` がインクルードされるインデックスファイルのURLを `c3_url` パラメータにより明示的に指定することができます。



``` yaml
coverage:
    # url of file which includes c3 router.
    c3_url: 'http://127.0.0.1:8000/index-test.php/'
```


`c3.php` がアプリケーションにインクルードされたのち、カバレッジの収集を開始することができます。
ローカルでアプリケーションを実行する場合に変更となる設定はありません。
すべてのコードカバレッジレポートは普段どおり収集され、その後マージされます。
考えてみてください、Codeceptionはローカルと同じようにリモートカバレッジを実行できるのです。

### リモートサーバー

もしテストを異なるサーバー上で実行している場合（もしくはウェブサーバーがカレントディレクトリのコードを使用していない場合）、設定ファイルに `remote` オプションを追加してください。
たとえば、受け入れテストでのリモートカバレッジを `acceptance.suite.yml` にて有効にしてみます。

``` yaml
coverage:
    remote: true
```

このオプションが有効となった場合、リモートコードカバレッジの結果はローカルのものとマージされません。
リモートとローカルのファイル群が同じパス上に存在する場合のみマージは行われます。
けれども、リモートサーバー上でテストを行う場合、同じパス上に存在するかどうかはわかりません。

リモートサーバーからのコードカバレッジの結果は `tests/_output` に保存されます。すでに述べた理由により、コンソールにはコードカバレッジの結果が表示されないことに注意してください。ローカルとリモートの結果はマージされませんし、コンソールにはローカルのコードカバレッジの結果が表示されます。

### リモートコンテキストオプション

HTMLレポートの生成にはデフォルトのタイムアウト30秒よりも少し時間がかかることがあります。また、SSLの設定を変更したい場合があるかもしれません（たとえばveryfy_peerオプション）
c3がウェブサーバー（ローカル、リモートのどちらであろうとも）に対してサービス要求をする方法を変更するには、`coverage` 設定の remote_context_options` キーを使用します。

``` yaml
coverage:
    remote_context_options:
        http:
            timeout: 60
        ssl:
            verify_peer: false
```

コンテキストオプションについては[php.netのドキュメント](http://php.net/manual/ja/context.php)に良くまとまっています。

## まとめ

ローカルとリモートのコードカバレッジ設定はこれまでになく簡単です。たった1つの設定を行い、1ファイルをインクルードするだけです！
**Codeceptionを使えばSeleniumテストのコードカバレッジレポートを簡単に生成することができます** （他の受け入れテストやAPIテストについても）。`acceptance`、`functional`、そして `unit` の各スイートが統合されたレポートは、アプリケーションのどの箇所がテストされていて、どの箇所がテストされていないのか、最も完全な情報を提供します。
